# FILE: widgetlog.tcl
#
# Each instance of the em_class net_widgetlog is a write-only record of 
# text messages, displayed in a Tk widget.
#
# A good candidate for inheritance.
#

# Only define em_widgetlog once (bootstrap protection)
if {[string match {} [info procs em_widgetlog]]} {

em_class em_widgetlog {
    public variable title {
        wm title [winfo toplevel $basewinpath] $title
    }
    public variable memory = 1000
    public variable dateformat = "\[%Y %b %d, %H:%M:%S %Z\] "
    public variable basewinpath = .
    const private variable winpath
    const private variable tw
    const private variable xscroll
    const private variable yscroll
    private variable lineslogged
    private variable linesinwidget

    private variable bindtags

    constructor {args} {
        set bindtags [list $this $class]
        eval $this configure $args
        # Don't add '.' suffix if there already is one.
        # (This happens chiefly when basewinpath is exactly '.')
        if {![string match {.} \
                [string index $basewinpath \
                [expr [string length $basewinpath]-1]]]} {
            append basewinpath {.}
        }

        set winpath ${basewinpath}w$this
        frame $winpath -bd 4 -relief ridge
        set tw [text $winpath.text -background white -height 5]
        set xscroll [scrollbar $winpath.xscroll -orient horizontal \
            -command [list $tw xview]]
        set yscroll [scrollbar $winpath.yscroll -orient vertical \
            -command [list $tw yview]]
        $tw configure -yscrollcommand [list $yscroll set] \
            -xscrollcommand [list $xscroll set] -state disabled -wrap none
        eval $this configure $args
        if {![info exists title]} {
            $this configure -title "Widget Log"
        }
        set lineslogged 0
        set linesinwidget 0
        pack $yscroll -side right -fill y
        pack $xscroll -side bottom -fill x
        pack $tw -side left -fill both -expand 1
        bind $winpath <Destroy> [list $this WinDestroy %W]
        pack $winpath -side bottom -fill both -expand 1
        $this Append "Starting $title ..."
    }

    method Append { args } { tw dateformat lineslogged linesinwidget memory } {
        # Cleanup routines might try to log messages after the window
        # is destroyed.  Prevent that error.
	if {![winfo exists $tw]} {
            return
        }
        $tw configure -state normal
        if {[llength $args] == 1} {
            set msg [lindex $args 0]
            set nonewline 0
        } elseif {([llength $args] == 2) \
                && [string match -nonewline [lindex $args 0]]} {
            set msg [lindex $args 1]
            set nonewline 1
        } else {
            error "usage: <em_widgetlog instance> Append ?-nonewline? <msg>"
        }
        # Workaround of bug in clock command in Tcl 7.5 - 7.6p2
        if {[regexp % $dateformat]} {
            set msg "[clock format [clock seconds] -format $dateformat]$msg"
        } else {
            set msg $dateformat$msg
        }
        set linestolog [llength [split $msg "\n"]]
        if {$nonewline} {
            $tw insert end "$msg"
        } else {
            $tw insert end "$msg\n"
        }
        $tw see end
        update idletasks
        incr lineslogged $linestolog
        incr linesinwidget $linestolog
        set excess [expr $linesinwidget - $memory]
        if {$excess > 0} {
            $this Purge $excess
        }
        $tw configure -state disabled
    }

    private method Purge { numlines } { tw linesinwidget } {
        $tw delete 1.0 "1.0 + $numlines lines"
        incr linesinwidget -$numlines
    }

    method HasGui {} {
        return 0
    }

    method WinDestroy { w } { winpath } {
        if {[string compare $winpath $w]} {
            return	;# Child destroy event
        }
        $this delete
    }

    # Append a line to this log
    destructor {
        em_eventHandler Generate delete $bindtags
	if {[info exists winpath] && [winfo exists $winpath]} {
            bind $winpath <Destroy> {}
            destroy $winpath
        }
    }

}

}
