# FILE: host.tcl
#
# An object which plays the role of a host within an interpreter.
#
# The Tcl event loop must be running for this event-driven object to function.
#

em_class net_host {

    # Index of host objects by name -- used to avoid duplicate objects.
    common index {[em_array new]}

    # The port on any host to which this object sends its commands
    common port 15136

    # Milliseconds to wait for host thread to start on localhost
    public variable startWait = 60000
    # The ID of the current timeout (returned from 'after' command)
    private variable timeoutevent

    # The name of the host this object acts as a proxy for
    const public variable hostname = localhost
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete \
                [list $this FatalError "Log dying"]]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

    # The connection through which we communicate to the host.
    private variable connection

    private variable msgHandler

    # Socket for receiving ping from localhost OOMMF host server
    private variable listensocket

    # Send id
    private variable id = 0

    # Reply message
    private variable reply

    # Is object ready to relay messages to and from the host?
    private variable ready = 0

    # Event handling support
    private variable bindtags
    private variable myEventHandlers = {}
    private variable rhandlers
    private variable toe

    private variable protocolversion

    # Do we display a gui?
    public variable gui = 0 {
        if {$gui} {
            if {[info exists myWidget]} {
                return
            }
            package require Tk
            set myWidget [net_hostGui new $parentWin -host $this]
            pack [$myWidget cget -winpath] -side top -fill both -expand 1
        } else {
            if {![info exists myWidget]} {
                return
            }
            $myWidget delete
            unset myWidget
        }
        upvar #0 ${this}_BtnVar btnvar
        set btnvar $gui
    }
    public variable parentWin = .
    private variable myWidget

    method ToggleGui { pwin } {
        set parentWin $pwin
        upvar #0 ${this}_BtnVar btnvar
        $this configure -gui $btnvar
    }

    # Constructor is a non-blocking operation.  After it returns an instance
    # id, the Ready method of that instance will return a boolean indicating
    # whether the object is ready.  If not, then when it becomes ready, it
    # will generate an event with id 'Ready'.  If the object cannot become
    # ready, it will delete itself, generating an event with is 'delete'.
    # Users of instances of this class should set up to handle those two
    # events, when the Ready method returns false.
    #
    constructor { args } {
        eval $this configure $args
        # Enforce only one object per host
        if {![catch {$index lookup $hostname} existinghost]} {
            unset hostname
            $this delete
            array set argarray $args
            catch {unset argarray(-hostname)}
            eval $existinghost configure [array get argarry]
            return $existinghost
        }
        $index set [list $hostname $this]
        if {![info exists log]} {
            # Provide a default log
            $this configure -log [em_log new -title "Host $hostname Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        after 0 $this Connect
    }

    method Connect {args} {
        if {[llength $args]} {
            $log Append "Received ping!"
        }
        if {[catch {
                net_connection new -hostname $hostname -port $port -log $log
                } msg]} {
            $this ConnectionFail
        } else {
            set connection $msg
$log Append "net_host Connect set connection to $connection"
            lappend myEventHandlers [em_eventHandler new $connection delete \
                    [list $this ConnectionFail]]
            lappend myEventHandlers [em_eventHandler new $connection Ready \
                    [list $this ConnectionReady]]
        }
        # NOTE: It is important that the ping channel be closed *after*
        # the new net_connection instance is created.  Otherwise Tcl may
        # try to re-use the same socket name, and if there are any bugs
        # in the Tcl library code which cleans up closed sockets (as it
        # appears there are, at least on some platforms), then there is the
        # danger of "cross-talk" between the closed ping socket and the new
        # connection socket.  In particular, we have observed spurious EOF
        # conditions appearing on the new net_connection's socket when the
        # order of these operations was reversed in an earlier version of this
        # software.  Better to avoid trouble by forcing Tcl to give us a 
        # socket with a new name.
        if {[llength $args]} {
            $log Append "Closing ping channel!"
            close [lindex $args 0]
        }
    }

    method ConnectionFail {} {
        if {$ready} {
            $this FatalError "Connection to $hostname died."
            return
        }
        if {[string match localhost $hostname] && ![info exists listensocket]} {
            if {[info exists connection]} {
                unset connection
            }
            $this StartLocalHost
        } else {
            $this FatalError "Can't open connection to $hostname"
        }
    }

    method ConnectionReady {} {
$log Append net_host:ConnectionReady
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this VerifyOOMMFHost]]
    }

    method VerifyOOMMFHost {} {
        if {[info exists listensocket]} {
            close $listensocket
            unset listensocket
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
            unset timeoutevent
        }
        $msgHandler delete
        unset msgHandler
        if {![regexp {OOMMF host protocol ([0-9.]+)} [$connection Get] \
                match version]} {
            # Not an OOMMF host
            $this FatalError "No OOMMF host server answering on $hostname"
            return
        }
        set protocolversion $version
        set rhandlers [em_array new]
        set toe [em_array new]
        $log Append "OOMMF host $hostname ready"
        set ready 1
        em_eventHandler Generate Ready $bindtags
    }

    method ProtocolVersion {} {
        if {!$ready} {
            error "host $hostname not ready"
        }
        return $protocolversion
    }

    # Launch an OOMMF host server on localhost
    private method StartLocalHost {} {
        set dir [[$class extension] directory]
        set hostthread [file join $dir threads host.tcl]
        if {[file readable $hostthread]} {
            # Set up to receive ping from host thread or timeout
            set listensocket [socket -server [list $this Connect] \
		    -myaddr localhost 0]  ;# Force to localhost for now
            set listenport [lindex [fconfigure $listensocket -sockname] 2]
            set timeoutevent [after $startWait $this StartTimeout]
            $log Append "Starting OOMMF host server on localhost"
            # Ought to redirect std channels somewhere
            global oommf
            if {[info exists oommf(shell)]} {
$log Append "$oommf(shell) -notk -nofork $hostthread $port $listenport >>& [em_extension NullFile] &"
                exec $oommf(shell) -notk -nofork $hostthread $port $listenport \
                       < [em_extension NullFile] >>& [em_extension NullFile] &
            } else {
                exec [info nameofexecutable] $hostthread $port $listenport \
                        < [em_extension NullFile] >>& [em_extension NullFile] &
            }
        } else {
            error "Installation error: No OOMMF host server available to start"
        }
    }

    method StartTimeout {} {
        $this FatalError \
                "Timed out waiting for OOMMF host server to start on localhost"
    }

    method Receive { rid } {  
        [$rhandlers lookup $rid] delete
        $rhandlers unset $rid
        [$toe lookup $rid] delete
        $toe unset $rid
        set reply [$connection Get] 
        em_eventHandler Generate Reply$rid $bindtags
    }

    method Timeout { rid } {
        [$rhandlers lookup $rid] delete
        $rhandlers unset $rid
        [$toe lookup $rid] delete
        $toe unset $rid
        $log Append "Timed out waiting for host $hostname reply to query $rid"
        em_eventHandler Generate Timeout$rid $bindtags
    }

    method Get {} {
        return $reply
    }

    # Send a command to the host
    method Send { args } {
        if {!$ready} {
            error "host $hostname not ready"
        }
        incr id
        set qid [eval $connection Query $args]
        $rhandlers set [list $id [em_eventHandler new $connection Reply$qid \
                [list $this Receive $id]]]
        $toe set [list $id [em_eventHandler new $connection Timeout$qid \
                [list $this Timeout $id]]]
        return $id
    }

    method Ready {} {
        return $ready
    }

    method FatalError { msg } {
        catch {$log Append $msg}
        catch {OMF_ModalDialog "Error in $class instance $this:" error $msg \
                {} {} 0}
        $this delete
    }

    destructor {
        set ready 0
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        if {[info exists myEventHandlers]} {
        foreach handler $myEventHandlers {
            $handler delete
        }
        }
        if {[info exists rhandlers]} {
            foreach idx [$rhandlers names] {
                [$rhandlers lookup $idx] delete
            }
            $rhandlers delete
        }
        if {[info exists toe]} {
            foreach idx [$toe names] {
                [$toe lookup $idx] delete
            }
            $toe delete
        }
        if {[info exists msgHandler]} {
            $msgHandler delete
        }
        if {[info exists connection]} {
$log Append "net_host delete destroying connection $connection"
            $connection delete
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
        }
        if {[info exists listensocket]} {
            close $listensocket
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this configure -log em_nop
        }
        if {[info exists myWidget]} {
            $this configure -gui 0
        }
        if {[info exists hostname]} {
            $index unset $hostname
        }
    }

}

