# FILE: mmDataTable.tcl
#
# Display Table of data values
#

set oommf(progname) "mmDataTable"

proc bgerror { msg } {
    puts stderr $msg
    flush stderr
}

# This code can be source'd from inside another script by setting
# no_net and no_event_loop as desired before sourcing this file.
# If no_net is set true, use the proc "AddTriples { triples }" to add
# data.

if {![info exists no_net]} {
    if {[string match "--nonet" [lindex $argv 0]]} {
	set no_net 1
    } else {
	set no_net 0      ;# Default is use net
    }
}

if {![info exists no_event_loop]} {
    set no_event_loop 0   ;# Default is to enter the event loop
}

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]
if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require Tk
package require em
if {!$no_net} { package require net }

# Death
proc Die { win } {
    if {![string match "." $win]} { return }  ;# Child destroy event
    global protocol
    if {[info exists protocol]} {
	$protocol delete; after 2000 exit
    } else {
	exit
    }
}
bind . <Destroy> {+Die %W}

# Menu bar
set menubar [frame .mb -borderwidth 2 -relief raised]

set fmb [menubutton .fmb -text File -menu .fmb.menu -underline 0]
set fmenu [menu .fmb.menu -tearoff 0]
$fmenu add command -label "Exit" -command { Die . } -underline 1
pack $fmb -side left -in $menubar

set dmb [menubutton .dmb -text Data -menu .dmb.menu -underline 0]
set dmenu [menu .dmb.menu -tearoff 1]
pack $dmb -side left -in $menubar

set omb [menubutton .omb -text Options -menu .omb.menu -underline 0]
set omenu [menu .omb.menu -tearoff 0]
pack $omb -side left -in $menubar
$omenu add command -label "Format" -underline 0 \
	-command { ChangeFormatInteractive }

set hmb [menubutton .hmb -text Help -menu .hmb.menu -underline 0 \
	-state disabled]
set hmenu [menu .hmb.menu -tearoff 0]
pack $hmb -side right -in $menubar


pack $menubar -side top -fill x
set panel [frame .panel -relief groove]
pack $panel -side top -fill x -expand 1 -anchor nw
grid columnconfigure $panel 0 -weight 0
grid columnconfigure $panel 1 -weight 0
grid columnconfigure $panel 2 -weight 1

proc DisplayListAppend { label foo } {
    global panel unit onoff fmt value fmtvalue
    if {$onoff($label)} {
	if {![winfo exists $panel.label$foo]} {
	    if {[string length $unit($label)]} {
		set txt "$label ($unit($label))"
	    } else {
		set txt "$label"
	    }
	    label $panel.label$foo -text $txt -anchor e
	}
	if {![winfo exists $panel.sep$foo]} {
	    label $panel.sep$foo -text ":"
	}
	if {![winfo exists $panel.value$foo]} {
	    set fmtvalue($label) [format $fmt($label) $value($label)]
	    label $panel.value$foo -textvariable fmtvalue($label) -anchor w
	    bind $panel.value$foo <Button-1> \
		    [list ChangeFormatInteractive $label]
	}
	grid $panel.label$foo $panel.sep$foo $panel.value$foo
	grid configure $panel.label$foo -sticky e
	grid configure $panel.value$foo -sticky w
    } else {
	if {[winfo exists $panel.label$foo] && \
		[winfo exists $panel.value$foo]} {
	    grid forget $panel.label$foo $panel.value$foo $panel.sep$foo 
	    destroy $panel.label$foo $panel.value$foo $panel.sep$foo
	}
    }
}

proc ChangeFormat { newformat {labellist {}} } {
    # Resets default format iff labellist is an empty list
    global fmt value fmtvalue onoff
    if {[llength $labellist]<1} {
	set lbllist [array names value]
	set fmt(default) $newformat
    } else {
	set lbllist $labellist
    }
    foreach lbl $lbllist {
	set fmt($lbl) $newformat
	if {[info exists onoff($lbl)] && $onoff($lbl) && \
	    [info exists value($lbl)] } {
	    set fmtvalue($lbl) [format $newformat $value($lbl)]
	}
    }
}

proc ChangeFormatInteractive { {label {}} } {
    global oommf fmt
    if {![string match {} $label] && [info exists fmt($label)]} {
	set format $fmt($label)
	set msg "Enter format string for label $label:"
	set labellist [list $label]
    } else {
	set format $fmt(default)
	set msg "Enter default numeric format string:"
	set labellist {}
    }
    set result [OMF_PromptDialog "$oommf(progname): Edit Format" \
	    $msg $format]
    if { [lindex $result 1] != 0 } { return }
    ChangeFormat [lindex $result 0] $labellist
}

######################################################################
# PROC: AddTriples
#
# Data import routine for mmDataTable widget
# Parameters:
#  Each input consists of a triple of values: label, unit, value
# Data storage:
#  The first element of each import triple is used as an index into
# the following collection of global arrays:
#
#     unit($label)     :  Value units
#     fmt($label)      :  Format code to use for display
#     value($label)    :  Value
#     fmtvalue($label) :  Formatted value
#     onoff($label)    :  Whether or not value is currently displayed
#
#  The fmt global array also includes a "default" element which is the
# the format applied to new labels.
#
set fmt(default) "%g"  ;# Default format
proc AddTriples { triples } {
    global unit fmt value fmtvalue dmenu onoff
    foreach trip $triples {
        set lbl [lindex $trip 0]
	set value($lbl) [lindex $trip 2]
        if {![info exists unit($lbl)]} {
            set unit($lbl) [lindex $trip 1]
            set onoff($lbl) 0
	    set fmt($lbl) $fmt(default)
            regsub -all "\[\r\n\t\]" $lbl " " txt
	    regsub -all "\[ \r\t\n\]" $lbl {} foo
            $dmenu add check -label $txt \
                    -variable onoff($lbl) \
                    -command [list DisplayListAppend $lbl $foo]
        }
	if {$onoff($lbl)} {
	    set fmtvalue($lbl) [format $fmt($lbl) $value($lbl)]
	}
    }
}

if {!$no_net} {
    set protocol [net_protocol new -name "OOMMF DataTable protocol 0.0"]
    $protocol AddMessage start HasGui {} {
	return [list start [list 0 0]]
    }
    $protocol AddMessage start DataTable { triples } {
	AddTriples $triples
	return [list start [list 0 0]]
    }
    set server [net_server new -protocol $protocol \
	    -alias mmDataTable -log em_nop]
    $server Start 0
}

wm title . mmDataTable<[pid]>

if {!$no_event_loop} { vwait omf_forever }
